import dotenv from 'dotenv';
import { defaultConfig } from './config/config';
import { createStats } from './config/stats';
import { VisionProviderFactory } from './providers/vision/visionProviderFactory';
import { TTSProviderFactory } from './providers/tts/ttsProviderFactory';
import { generateAudioDescription } from './utils/processor';
import { estimateCost } from './utils/costEstimator';
import { loadConfigFromFile, saveConfigToFile } from './utils/configUtils';
import { parseCommandLineArgs } from './cli/args';
import { ProcessingResult, CostBreakdown } from './interfaces';

// Load environment variables
dotenv.config();

// Main execution when run directly
async function main(): Promise<void> {
  // Parse command line arguments
  const argv = parseCommandLineArgs();

  // Start with default config
  let config = { ...defaultConfig };

  // If a config file is specified, load it
  if (argv.config) {
    const fileConfig = loadConfigFromFile(argv.config);
    config = { ...config, ...fileConfig };
  }

  // Override with any command line arguments
  Object.keys(argv).forEach(key => {
    if (key !== '_' && key !== '$0' && key !== 'config' && key !== 'saveConfig' &&
      key !== 'estimate' && key !== 'help' && key !== 'version' &&
      argv[key as keyof typeof argv] !== undefined) {
      (config as any)[key] = argv[key as keyof typeof argv];
    }
  });

  // Handle nested provider configurations
  if (argv.visionModel) {
    if (!config.visionProviders[config.visionProvider]) {
      config.visionProviders[config.visionProvider] = { model: '' };
    }
    config.visionProviders[config.visionProvider].model = argv.visionModel as string;
  }

  if (argv.ttsModel) {
    if (!config.ttsProviders[config.ttsProvider]) {
      config.ttsProviders[config.ttsProvider] = { model: '' };
    }
    config.ttsProviders[config.ttsProvider].model = argv.ttsModel as string;
  }

  if (argv.ttsVoice) {
    if (!config.ttsProviders[config.ttsProvider]) {
      config.ttsProviders[config.ttsProvider] = { model: '', voice: '' };
    }
    config.ttsProviders[config.ttsProvider].voice = argv.ttsVoice as string;
  }

  // Save configuration if requested
  if (argv.saveConfig) {
    saveConfigToFile(argv.saveConfig as string, config);
  }

  // Check if a video file is provided
  if (argv._.length < 1) {
    console.error('Error: No video file specified');
    console.log('Usage: node script.js <video_file_path> [options]');
    console.log('Use --help for more information');
    process.exit(1);
  }

  const videoFilePath = argv._[0] as string;

  // Run estimation or full processing
  if (argv.estimate) {
    try {
      const costBreakdown = await estimateCost(videoFilePath, config);
      console.log('\n=== COST ESTIMATION ===');
      console.log(JSON.stringify(costBreakdown, null, 2));
      console.log(`\nEstimated total cost: ${costBreakdown.apiCosts.total}`);
      console.log(`Estimated processing time: ${costBreakdown.estimates.estimatedProcessingTimeMinutes.toFixed(1)} minutes`);
      console.log('Note: Actual costs may vary based on image complexity and actual response lengths.');
    } catch (err) {
      console.error('Error estimating costs:', err);
    }
  } else {
    // Run the full generator
    try {
      // Initialize the stats object
      const stats = createStats();
      
      // Initialize providers
      const visionProvider = VisionProviderFactory.getProvider(config);
      const ttsProvider = TTSProviderFactory.getProvider(config);
      
      await generateAudioDescription(videoFilePath, visionProvider, ttsProvider, config, stats);
    } catch (err) {
      console.error('Error generating audio description:', err);
    }
  }
}

// Only run the main function if this file is executed directly (not imported)
if (require.main === module) {
  main().catch(err => console.error('Unhandled error:', err));
}

// Export functions and types for use as a module
export {
  generateAudioDescription,
  estimateCost,
  defaultConfig,
  VisionProviderFactory,
  TTSProviderFactory,
  ProcessingResult,
  CostBreakdown
};